@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico.
-   Copyright (C) 1998-2018 Sergey Poznyakoff
+   Copyright (C) 1998-2020 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -29,7 +29,7 @@ dico_list_t sasl_anon_groups;
 static Gsasl *ctx;   
 
 static int
-cmp_names(const void *item, void *data)
+cmp_names(const void *item, const void *data, void *closure)
 {
     return c_strcasecmp(item, data);
 }
@@ -38,9 +38,9 @@ static int
 disabled_mechanism_p(char *name)
 {
     if (sasl_enabled_mech
-	&& !_dico_list_locate(sasl_enabled_mech, name, cmp_names))
+	&& !_dico_list_locate(sasl_enabled_mech, name, cmp_names, NULL))
 	return 1;
-    return !!_dico_list_locate(sasl_disabled_mech, name, cmp_names);
+    return !!_dico_list_locate(sasl_disabled_mech, name, cmp_names, NULL);
 }
 
 static void
@@ -326,7 +326,7 @@ callback(Gsasl *ctx, Gsasl_session *sctx, Gsasl_property prop)
 }
 
 static int
-init_sasl_0()
+init_sasl_0(void)
 {
     int rc = gsasl_init(&ctx);
     if (rc != GSASL_OK) {
@@ -339,7 +339,7 @@ init_sasl_0()
 }
 
 static int
-init_sasl_1()
+init_sasl_1(void *data)
 {
     static struct dicod_command cmd =
 	{ "SASLAUTH", 2, 3, "mechanism [initial-response]",
@@ -357,7 +357,7 @@ init_sasl_1()
 }
 
 void
-register_sasl()
+register_sasl(void)
 {
   int rc;
   char *listmech;
@@ -390,7 +390,7 @@ register_sasl()
 
 #else
 void
-register_sasl()
+register_sasl(void)
 {
     /* nothing */
 }
