@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico.
-   Copyright (C) 1998-2000, 2008, 2010, 2012 Sergey Poznyakoff
+   Copyright (C) 1998-2020 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -24,7 +24,7 @@ int single_process; /* Single process mode */
 char *config_file = SYSCONFIG "/dicod.conf" ;
 int config_lint_option; /* Check configuration file syntax and exit. */
 /* Location of the pidfile */
-char *pidfile_name = LOCALSTATEDIR "/run/dicod.pid";
+char *pidfile_name = DEFAULT_PIDFILE_NAME;
 
 /* Operation mode */
 int mode = MODE_DAEMON;
@@ -42,9 +42,9 @@ unsigned int shutdown_timeout = 5;
 /* Inactivity timeout */
 unsigned int inactivity_timeout = 0;
 
-/* Syslog parameters: */ 
+/* Syslog parameters: */
 int log_to_stderr;  /* Log to stderr */
-const char *log_tag; 
+const char *log_tag;
 int log_facility = LOG_FACILITY;
 int log_print_severity;
 int transcript;
@@ -64,7 +64,7 @@ char *initial_banner_text;
 char *help_text;
 
 /* List of sockets to listen on for the requests */
-dico_list_t /* of struct sockaddr */ listen_addr;
+dico_list_t /* of struct grecs_sockaddr */ listen_addr;
 
 /* User database for AUTH */
 dico_udb_t user_db;
@@ -125,7 +125,7 @@ grecs_list_iterate(struct grecs_list *list, grecs_list_iterator_t fun,
 	    break;
     }
 }
-	
+
 
 
 /* Configuration */
@@ -182,7 +182,7 @@ cb_dico_sockaddr_list(enum grecs_callback_command cmd,
 {
     dico_list_t *plist = varptr, list;
     struct grecs_sockaddr *sp;
-    
+
     if (*plist)
 	list = *plist;
     else {
@@ -190,16 +190,18 @@ cb_dico_sockaddr_list(enum grecs_callback_command cmd,
 //FIXME	dico_list_set_free_item(list, dicod_free_item, NULL);
 	*plist = list;
     }
-    
+
     switch (value->type) {
     case GRECS_TYPE_STRING:
-	sp = xmalloc(sizeof(*sp));
-	if (grecs_string_convert(sp, grecs_type_sockaddr,
-				 value->v.string, locus)) {
-	    free(sp);
+	if (grecs_string_convert(&sp, grecs_type_sockaddr,
+				 value->v.string, locus))
 	    return 1;
+	while (sp) {
+	    struct grecs_sockaddr *next = sp->next;
+	    sp->next = NULL;
+	    xdico_list_append(list, sp);
+	    sp = next;
 	}
-	xdico_list_append(list, sp);
 	break;
 
     case GRECS_TYPE_LIST: {
@@ -213,13 +215,15 @@ cb_dico_sockaddr_list(enum grecs_callback_command cmd,
 			    _("list element must be a socket specification"));
 		return 1;
 	    }
-	    sp = xmalloc(sizeof(*sp));
-	    if (grecs_string_convert(sp, grecs_type_sockaddr,
-				     vp->v.string, &vp->locus)) {
-		free(sp);
+	    if (grecs_string_convert(&sp, grecs_type_sockaddr,
+				     vp->v.string, &vp->locus))
 		return 1;
+	    while (sp) {
+		struct grecs_sockaddr *next = sp->next;
+		sp->next = NULL;
+		xdico_list_append(list, sp);
+		sp = next;
 	    }
-	    xdico_list_append(list, sp);
 	}
 	break;
     }
@@ -231,7 +235,6 @@ cb_dico_sockaddr_list(enum grecs_callback_command cmd,
     return 0;
 }
 
-
 int
 allow_cb(enum grecs_callback_command cmd,
 	 grecs_locus_t *locus,
@@ -274,10 +277,10 @@ acl_cb(enum grecs_callback_command cmd,
 {
     void **pdata = cb_data;
     dicod_acl_t acl;
-    
+
     switch (cmd) {
     case grecs_callback_section_begin:
-	if (value->type != GRECS_TYPE_STRING) 
+	if (value->type != GRECS_TYPE_STRING)
 	    grecs_error(locus, 0, _("ACL name must be a string"));
 	else if (!value->v.string)
 	    grecs_error(locus, 0, _("missing ACL name"));
@@ -299,7 +302,7 @@ acl_cb(enum grecs_callback_command cmd,
     case grecs_callback_section_end:
     case grecs_callback_set_value:
 	break;
-    }	
+    }
     return 0;
 }
 
@@ -333,7 +336,7 @@ apply_acl_cb(enum grecs_callback_command cmd,
 	grecs_error(locus, 0, _("expected scalar value"));
 	return 1;
     }
-    if ((*pacl =  dicod_acl_lookup(value->v.string)) == NULL) {
+    if ((*pacl = dicod_acl_lookup(value->v.string)) == NULL) {
 	grecs_error(locus, 0, _("no such ACL: `%s'"), value->v.string);
 	return 1;
     }
@@ -349,7 +352,8 @@ set_user(enum grecs_callback_command cmd,
 	 void *cb_data)
 {
     struct passwd *pw;
-    
+    char *s;
+
     if (cmd != grecs_callback_set_value) {
 	grecs_error(locus, 0, _("Unexpected block statement"));
 	return 1;
@@ -359,8 +363,17 @@ set_user(enum grecs_callback_command cmd,
 	grecs_error(locus, 0, _("expected scalar value but found list"));
 	return 1;
     }
-    
-    pw = getpwnam(value->v.string);
+    s = value->v.string;
+    if (*s == '+') {
+	char *q;
+	unsigned long n = strtoul(s + 1, &q, 0);
+	if (*q) {
+	    grecs_error(locus, 0, _("not a valid UID number: %s"), s);
+	    return 1;
+	}
+	pw = getpwuid(n);
+    } else
+	pw = getpwnam(s);
     if (!pw) {
 	grecs_error(locus, 0, _("%s: no such user"), value->v.string);
 	return 1;
@@ -382,7 +395,7 @@ set_supp_group_iter(grecs_value_t *value, void *data)
     return set_supp_group(grecs_callback_set_value,
 			  &value->locus, NULL, value, NULL);
 }
-	
+
 static int
 set_supp_group(enum grecs_callback_command cmd,
 	       grecs_locus_t *locus,
@@ -395,19 +408,37 @@ set_supp_group(enum grecs_callback_command cmd,
 	return 1;
     }
 
-    if (!group_list)
+    if (!group_list) {
 	group_list = xdico_list_create();
-    
+	dico_list_set_free_item(group_list, dicod_free_item, NULL);
+    }
+
     if (value->type == GRECS_TYPE_LIST)
 	grecs_list_iterate(value->v.list, set_supp_group_iter, NULL);
     else {
-	struct group *group = getgrnam(value->v.string);
-	if (group)
-	    xdico_list_append(group_list, (void*)group->gr_gid);
-	else {
-	    grecs_error(locus, 0, _("%s: unknown group"), value->v.string);
-	    return 1;
+	char *s = value->v.string;
+	gid_t gid, *gp;
+
+	if (*s == '+') {
+	    char *q;
+	    unsigned long n = strtoul(s + 1, &q, 0);
+	    if (*q) {
+		grecs_error(locus, 0, _("not a valid GID number: %s"), s);
+		return 1;
+	    }
+	    gid = n;
+	} else {
+	    struct group *group = getgrnam(s);
+	    if (!group) {
+		grecs_error(locus, 0, _("%s: unknown group"), value->v.string);
+		return 1;
+	    }
+	    gid = group->gr_gid;
 	}
+
+	gp = xmalloc(sizeof(*gp));
+	*gp = gid;
+	xdico_list_append(group_list, gp);
     }
     return 0;
 }
@@ -424,7 +455,7 @@ set_mode(enum grecs_callback_command cmd,
 	{ "inetd", MODE_INETD },
 	{ NULL }
     };
-	
+
     if (cmd != grecs_callback_set_value) {
 	grecs_error(locus, 0, _("Unexpected block statement"));
 	return 1;
@@ -442,7 +473,7 @@ set_mode(enum grecs_callback_command cmd,
 }
 
 static struct xlat_tab syslog_facility_tab[] = {
-    { "USER",    LOG_USER },   
+    { "USER",    LOG_USER },
     { "DAEMON",  LOG_DAEMON },
     { "AUTH",    LOG_AUTH },
     { "AUTHPRIV",LOG_AUTHPRIV },
@@ -487,7 +518,7 @@ set_log_facility(enum grecs_callback_command cmd,
 }
 
 static int
-cmp_modinst_ident(const void *item, void *data)
+cmp_modinst_ident(const void *item, const void *data, void *closure)
 {
     const dicod_module_instance_t *inst = item;
     if (!inst->ident)
@@ -498,9 +529,9 @@ cmp_modinst_ident(const void *item, void *data)
 static int
 _add_simple_module(grecs_value_t *value, void *unused_data)
 {
-    if (value->type != GRECS_TYPE_STRING) 
+    if (value->type != GRECS_TYPE_STRING)
 	grecs_error(&value->locus, 0, _("tag must be a string"));
-    else if (value->v.string == NULL) 
+    else if (value->v.string == NULL)
 	grecs_error(&value->locus, 0, _("missing tag"));
     else {
 	dicod_module_instance_t *inst = xzalloc(sizeof(*inst));
@@ -520,30 +551,25 @@ load_module_cb(enum grecs_callback_command cmd,
 {
     dicod_module_instance_t *inst;
     void **pdata = cb_data;
-    
-    if (!modinst_list) {
-	modinst_list = xdico_list_create();
-	dico_list_set_comparator(modinst_list, cmp_modinst_ident);
-    }
-    
+
     switch (cmd) {
     case grecs_callback_section_begin:
 	inst = xzalloc(sizeof(*inst));
-	if (value->type != GRECS_TYPE_STRING) 
+	if (value->type != GRECS_TYPE_STRING)
 	    grecs_error(locus, 0, _("tag must be a string"));
-	else if (value->v.string == NULL) 
+	else if (value->v.string == NULL)
 	    grecs_error(locus, 0, _("missing tag"));
 	else
 	    inst->ident = xstrdup(value->v.string);
 	*pdata = inst;
 	break;
-	
+
     case grecs_callback_section_end:
 	inst = *pdata;
 	xdico_list_append(modinst_list, inst);
 	*pdata = NULL;
 	break;
-	
+
     case grecs_callback_set_value:
 	switch (value->type) {
 	case GRECS_TYPE_STRING:
@@ -562,18 +588,18 @@ load_module_cb(enum grecs_callback_command cmd,
 }
 
 static int
-cmp_database_name(const void *item, void *data)
+cmp_database_name(const void *item, const void *data, void *closure)
 {
     const dicod_database_t *db = item;
     int rc;
-    
+
     if (!db->name)
 	return 1;
     rc = strcmp(db->name, (const char*)data);
-    if (rc == 0 && !database_visible_p(db))
+    if (rc == 0 && !database_is_visible(db))
 	rc = 1;
     return rc;
-}    
+}
 
 static int
 set_database(enum grecs_callback_command cmd,
@@ -584,13 +610,14 @@ set_database(enum grecs_callback_command cmd,
 {
     dicod_database_t *dict;
     void **pdata = cb_data;
-    
+
     switch (cmd) {
     case grecs_callback_section_begin:
 	dict = xzalloc(sizeof(*dict));
+	dict->visible = 1;
 	*pdata = dict;
 	break;
-	
+
     case grecs_callback_section_end:
 	dict = *pdata;
 	if (!dict->name) {
@@ -598,18 +625,25 @@ set_database(enum grecs_callback_command cmd,
 	    break;
 	}
 
+	if (!dict->instance) {
+	    grecs_error(locus, 0,
+			_("no handler defined for the database"));
+	    dicod_database_free(dict);
+	    return 1;
+	}
+	
 	if (!database_list) {
 	    database_list = xdico_list_create();
-	    dico_list_set_comparator (database_list, cmp_database_name);
+	    dico_list_set_comparator (database_list, cmp_database_name, NULL);
 	}
 	if (dict->langlist[0] || dict->langlist[1])
 	     /* Prevent dico_db_lang from being called */
-	    dict->flags |= DICOD_DBF_LANG;
-	
+	    dict->flags |= DICO_DBF_LANG;
+
 	xdico_list_append(database_list, dict);
 	*pdata = NULL;
 	break;
-	
+
     case grecs_callback_set_value:
 	grecs_error(locus, 0, _("invalid use of block statement"));
     }
@@ -625,9 +659,8 @@ set_dict_handler(enum grecs_callback_command cmd,
 {
     dicod_module_instance_t *inst;
     dicod_database_t *db = varptr;
-    int rc;
     struct wordsplit ws;
-    
+
     if (cmd != grecs_callback_set_value) {
 	grecs_error(locus, 0, _("Unexpected block statement"));
 	return 1;
@@ -639,11 +672,11 @@ set_dict_handler(enum grecs_callback_command cmd,
     }
 
     if (wordsplit(value->v.string, &ws, WRDSF_DEFFLAGS)) {
-	grecs_error(locus, rc, _("cannot parse command line `%s': %s"),
+	grecs_error(locus, 0, _("cannot parse command line `%s': %s"),
 		    value->v.string, wordsplit_strerror (&ws));
-	dicod_database_free(db); 
+	dicod_database_free(db);
 	return 1;
-    } 
+    }
 
     db->argc = ws.ws_wordc;
     db->argv = ws.ws_wordv;
@@ -660,7 +693,7 @@ set_dict_handler(enum grecs_callback_command cmd,
 	return 1;
     }
     db->instance = inst;
-    
+
     return 0;
 }
 
@@ -690,7 +723,7 @@ enable_capability(enum grecs_callback_command cmd,
     }
     if (value->type == GRECS_TYPE_LIST)
 	grecs_list_iterate(value->v.list, set_capability, locus);
-    else if (dicod_capa_add(value->v.string)) 
+    else if (dicod_capa_add(value->v.string))
 	grecs_error(locus, 0, _("unknown capability: %s"), value->v.string);
     return 0;
 }
@@ -704,7 +737,7 @@ mime_headers_cb (enum grecs_callback_command cmd,
 {
     dico_assoc_list_t *pasc = varptr;
     const char *enc;
-    
+
     if (cmd != grecs_callback_set_value) {
 	grecs_error(locus, 0, _("Unexpected block statement"));
 	return 1;
@@ -715,17 +748,17 @@ mime_headers_cb (enum grecs_callback_command cmd,
 	return 1;
     }
 
-    if (dico_header_parse(pasc, value->v.string)) 
+    if (dico_header_parse(pasc, value->v.string))
 	grecs_error(locus, 0, _("cannot parse headers: %s"),
 		     strerror(errno));
 
-    if (enc = dico_assoc_find(*pasc, CONTENT_TRANSFER_ENCODING_HEADER)) {
-        if (!(strcmp(enc, "quoted-printable") == 0
+    if ((enc = dico_assoc_find(*pasc, CONTENT_TRANSFER_ENCODING_HEADER))) {
+	if (!(strcmp(enc, "quoted-printable") == 0
 	      || strcmp(enc, "base64") == 0
 	      || strcmp(enc, "8bit") == 0))
 	    grecs_error(locus, 0, _("unknown encoding type: %s"), enc);
     }
-    
+
     return 0;
 }
 
@@ -736,6 +769,70 @@ struct grecs_keyword kwd_load_module[] = {
     { NULL }
 };
 
+int
+member_database_cb(enum grecs_callback_command cmd,
+		   grecs_locus_t *locus,
+		   void *varptr,
+		   grecs_value_t *value,
+		   void *cb_data)
+{
+    dico_list_t *listptr = varptr;
+    char const *dbname;
+    enum mime_cond cond = cond_any;
+    grecs_value_t *vp;
+    struct vdb_member *memb;
+
+    if (cmd != grecs_callback_set_value) {
+	grecs_error(locus, 0, _("Unexpected block statement"));
+	return 1;
+    }
+
+    switch (value->type) {
+    case GRECS_TYPE_ARRAY:
+	if (value->v.arg.c != 2) {
+	    grecs_error(locus, 0, _("too many arguments"));
+	    return 1;
+	}
+	vp = value->v.arg.v[0];
+	if (vp->type != GRECS_TYPE_STRING) {
+	    grecs_error(locus, 0, _("first argument must be string"));
+	    return 1;
+	}
+	dbname = vp->v.string;
+	vp = value->v.arg.v[1];
+	if (vp->type != GRECS_TYPE_STRING) {
+	    grecs_error(locus, 0, _("second argument must be string"));
+	    return 1;
+	}
+	if (strcmp(vp->v.string, "mime") == 0)
+	    cond = cond_mime;
+	else if (strcmp(vp->v.string, "nomime") == 0)
+	    cond = cond_nomime;
+	else {
+	    grecs_error(locus, 0, _("condition must be mime or nomime"));
+	    return 1;
+	}
+	break;
+
+    case GRECS_TYPE_STRING:
+	dbname = value->v.string;
+	break;
+
+    default:
+	grecs_error(locus, 0, _("expected scalar value"));
+	return 1;
+    }
+
+    if (!*listptr)
+	*listptr = vdb_list_create();
+
+    memb = xzalloc(sizeof(memb[0]));
+    memb->name = xstrdup(dbname);
+    memb->cond = cond;
+    xdico_list_append(*listptr, memb);
+    return 0;
+}
+
 struct grecs_keyword kwd_database[] = {
     { "name", N_("word"), N_("Dictionary name (a single word)."),
       grecs_type_string, GRECS_DFLT,
@@ -769,6 +866,18 @@ struct grecs_keyword kwd_database[] = {
       grecs_type_string, GRECS_DFLT,
       NULL, offsetof(dicod_database_t, mime_headers),
       mime_headers_cb },
+    { "visible", N_("bool"),
+      N_("Set database visibility"),
+      grecs_type_bool, GRECS_DFLT,
+      NULL, offsetof(dicod_database_t, visible) },
+    { "database", N_("<name: string> [mime|nomime]"),
+      N_("For virtual database only: name of the member database."
+	 " Optional mime|nomime specifies the condition under which"
+	 " the database is to be used."),
+      grecs_type_string, GRECS_DFLT,
+      NULL, offsetof(dicod_database_t, extra),
+      member_database_cb },
+
     { NULL }
 };
 
@@ -807,13 +916,13 @@ user_db_config(enum grecs_callback_command cmd,
 {
     struct user_db_conf *cfg = varptr;
     void **pdata = cb_data;
-    
+
     switch (cmd) {
     case grecs_callback_section_begin:
 	cfg->locus = *locus;
 	cfg->locus.beg.file = xstrdup(cfg->locus.beg.file);
 	cfg->locus.end.file = xstrdup(cfg->locus.end.file);
-	if (value->type != GRECS_TYPE_STRING) 
+	if (value->type != GRECS_TYPE_STRING)
 	    grecs_error(locus, 0, _("URL must be a string"));
 	else if (!value->v.string)
 	    grecs_error(locus, 0, _("empty URL"));
@@ -821,15 +930,15 @@ user_db_config(enum grecs_callback_command cmd,
 	    cfg->url = xstrdup(value->v.string);
 	*pdata = cfg;
 	break;
-	
+
     case grecs_callback_section_end:
 	break;
-	
+
     case grecs_callback_set_value:
 	cfg->locus = *locus;
 	cfg->locus.beg.file = xstrdup(cfg->locus.beg.file);
 	cfg->locus.end.file = xstrdup(cfg->locus.end.file);
-	if (value->type != GRECS_TYPE_STRING) 
+	if (value->type != GRECS_TYPE_STRING)
 	    grecs_error(locus, 0, _("URL must be a string"));
 	else if (!value->v.string)
 	    grecs_error(locus, 0, _("empty URL"));
@@ -839,7 +948,7 @@ user_db_config(enum grecs_callback_command cmd,
 }
 
 static void
-init_user_db()
+init_user_db(void)
 {
     if (!user_db_cfg.url)
 	sasl_enable = 0;
@@ -895,7 +1004,7 @@ sasl_cb(enum grecs_callback_command cmd,
 	void *cb_data)
 {
     if (cmd == grecs_callback_set_value) {
-	if (value->type != GRECS_TYPE_STRING) 
+	if (value->type != GRECS_TYPE_STRING)
 	    grecs_error(locus, 0, _("expected boolean value but found list"));
 	else
 	    grecs_string_convert(&sasl_enable, grecs_type_bool,
@@ -939,7 +1048,7 @@ flush_strat_forward_fn(void *item, void *data)
 }
 
 void
-flush_strat_forward()
+flush_strat_forward(void)
 {
     dico_strategy_iterate(flush_strat_forward_fn, NULL);
     dico_list_destroy(&strat_forward);
@@ -956,7 +1065,7 @@ strategy_cb(enum grecs_callback_command cmd,
 
     switch (cmd) {
     case grecs_callback_section_begin:
-	if (value->type != GRECS_TYPE_STRING) 
+	if (value->type != GRECS_TYPE_STRING)
 	    grecs_error(locus, 0, _("Section name must be a string"));
 	else if (!value->v.string)
 	    grecs_error(locus, 0, _("missing section name"));
@@ -969,19 +1078,20 @@ strategy_cb(enum grecs_callback_command cmd,
 		if (!strat_forward) {
 		    strat_forward = xdico_list_create();
 		    dico_list_set_comparator (strat_forward,
-					      dico_strat_name_cmp);
+					      dico_strat_name_cmp,
+					      NULL);
 		    dico_list_set_free_item(strat_forward,
 					    dico_strat_free, NULL);
 		}
 		xdico_list_append(strat_forward, strat);
-	    } 
+	    }
 	    *pdata = strat;
 	}
 	break;
-		
+
     case grecs_callback_section_end:
 	break;
-	
+
     case grecs_callback_set_value:
 	grecs_error(locus, 0, _("Unexpected statement"));
     }
@@ -996,7 +1106,7 @@ strategy_deny_all_cb(enum grecs_callback_command cmd,
 		     void *cb_data)
 {
     int bool;
-    
+
     if (cmd != grecs_callback_set_value) {
 	grecs_error(locus, 0, _("Unexpected block statement"));
 	return 1;
@@ -1063,7 +1173,7 @@ strategy_deny_length(enum grecs_callback_command cmd,
     }
     if (value->type == GRECS_TYPE_STRING) {
 	size_t val;
-	
+
 	if (grecs_string_convert(&val, grecs_type_size,
 				 value->v.string, locus))
 	    return 0;
@@ -1218,7 +1328,7 @@ struct grecs_keyword keywords[] = {
       N_("Maximum number of children running simultaneously."),
       grecs_type_uint, GRECS_DFLT, &max_children, 0 },
     { "log-tag", N_("arg"),  N_("Tag syslog diagnostics with this tag."),
-      grecs_type_string, GRECS_DFLT, &log_tag, 0 },
+      grecs_type_string, GRECS_DFLT|GRECS_CONST, &log_tag, 0 },
     { "log-facility", N_("arg"),
       N_("Set syslog facility. Arg is one of the following: user, daemon, "
 	 "auth, authpriv, mail, cron, local0 through local7 "
@@ -1229,15 +1339,15 @@ struct grecs_keyword keywords[] = {
       grecs_type_bool, GRECS_DFLT, &log_print_severity, 0 },
     { "access-log-format", N_("fmt"),
       N_("Set format string for access log file."),
-      grecs_type_string, GRECS_DFLT, &access_log_format, },
+      grecs_type_string, GRECS_DFLT|GRECS_CONST, &access_log_format, },
     { "access-log-file", N_("name"),
       N_("Set access log file name."),
-      grecs_type_string, GRECS_DFLT, &access_log_file },
+      grecs_type_string, GRECS_DFLT|GRECS_CONST, &access_log_file },
     { "transcript", N_("arg"), N_("Log session transcript."),
       grecs_type_bool, GRECS_DFLT, &transcript },
     { "pidfile", N_("name"),
       N_("Store PID of the master process in this file."),
-      grecs_type_string, GRECS_DFLT, &pidfile_name, },
+      grecs_type_string, GRECS_DFLT|GRECS_CONST, &pidfile_name, },
     { "shutdown-timeout", N_("seconds"),
       N_("Wait this number of seconds for all children to terminate."),
       grecs_type_uint, GRECS_DFLT, &shutdown_timeout },
@@ -1248,11 +1358,11 @@ struct grecs_keyword keywords[] = {
       grecs_type_sockaddr, GRECS_LIST, &listen_addr, 0, cb_dico_sockaddr_list },
     { "initial-banner-text", N_("text"),
       N_("Display this text in the initial 220 banner"),
-      grecs_type_string, GRECS_DFLT, &initial_banner_text },
+      grecs_type_string, GRECS_DFLT|GRECS_CONST, &initial_banner_text },
     { "help-text", N_("text"),
       N_("Display this text in reply to the HELP command. If text "
 	 "begins with a +, usual command summary is displayed before it."),
-      grecs_type_string, GRECS_DFLT, &help_text },
+      grecs_type_string, GRECS_DFLT|GRECS_CONST, &help_text },
     { "hostname", N_("name"), N_("Override the host name."),
       grecs_type_string, GRECS_DFLT, &hostname },
     { "capability", N_("arg"), N_("Request additional capabilities."),
@@ -1314,30 +1424,36 @@ struct grecs_keyword keywords[] = {
 };
 
 void
-config_init()
+config_init(void)
 {
     grecs_include_path_setup (DEFAULT_VERSION_INCLUDE_DIR,
 			      DEFAULT_INCLUDE_DIR, NULL);
     grecs_preprocessor = DEFAULT_PREPROCESSOR;
     grecs_log_to_stderr = 1;
     grecs_default_port = htons(DICO_DICT_PORT);
+    grecs_parser_options = GRECS_OPTION_QUOTED_STRING_CONCAT;
+
+    modinst_list = xdico_list_create();
+    dico_list_set_comparator(modinst_list, cmp_modinst_ident, NULL);
+
+    dicod_builtin_module_init();
 }
 
 void
-config_parse()
+config_parse(void)
 {
     struct grecs_node *tree;
 
-    tree = grecs_parse (config_file);
+    tree = grecs_parse(config_file);
     if (!tree)
-	exit (EX_CONFIG);
-    if (grecs_tree_process (tree, keywords) || grecs_error_count)
-	exit (EX_CONFIG);
-    grecs_tree_free (tree);
+	exit(EX_CONFIG);
+    if (grecs_tree_process(tree, keywords) || grecs_error_count)
+	exit(EX_CONFIG);
+    grecs_tree_free(tree);
 }
 
 void
-config_help()
+config_help(void)
 {
     static char docstring[] =
 	N_("Configuration file structure for dicod.\n"
@@ -1348,7 +1464,7 @@ config_help()
 
 
 int
-show_sys_info_p()
+show_sys_info_p(void)
 {
     if (!show_sys_info)
 	return 1;
@@ -1356,32 +1472,48 @@ show_sys_info_p()
 }
 
 void
-reset_db_visibility()
+reset_db_visibility(void)
 {
     dicod_database_t *db;
     dico_iterator_t itr;
 
     itr = xdico_list_iterator(database_list);
-    for (db = dico_iterator_first(itr); db; db = dico_iterator_next(itr)) 
-	db->visible = 1;
+    for (db = dico_iterator_first(itr); db; db = dico_iterator_next(itr))
+	db->session_visible = db->visible;
     dico_iterator_destroy(&itr);
 }
 
+int
+database_session_visibility(dicod_database_t *db)
+{
+    if (!dicod_acl_check(db->acl, dicod_acl_check(global_acl, 1)))
+	return 0;
+    else {
+	dico_list_t list[2];
+	dicod_database_get_languages(db, list);
+	if (!dicod_lang_check(list))
+	    return 0;
+    }
+    return 1;
+}
+
 void
-check_db_visibility()
+check_db_visibility(void)
 {
     dicod_database_t *db;
     dico_iterator_t itr;
     int global = dicod_acl_check(global_acl, 1);
-    
+
     itr = xdico_list_iterator(database_list);
     for (db = dico_iterator_first(itr); db; db = dico_iterator_next(itr)) {
-	if (!dicod_acl_check(db->acl, global))
-	    db->visible = 0;
+	if (!db->visible)
+	    db->session_visible = 0;
+	else if (!dicod_acl_check(db->acl, global))
+	    db->session_visible = 0;
 	else {
 	    dico_list_t list[2];
-	    dicod_get_database_languages(db, list);
-	    db->visible = dicod_lang_check(list);
+	    dicod_database_get_languages(db, list);
+	    db->session_visible = dicod_lang_check(list);
 	}
     }
     dico_iterator_destroy(&itr);
@@ -1399,13 +1531,13 @@ _count_databases(void *item, void *data)
 {
     const dicod_database_t *db = item;
     size_t *pcount = data;
-    if (database_visible_p(db))
+    if (database_is_visible(db))
 	++*pcount;
     return 0;
 }
 
 size_t
-database_count()
+database_count(void)
 {
     size_t count = 0;
     dico_list_iterate(database_list, _count_databases, &count);
@@ -1418,10 +1550,10 @@ database_iterate(dico_list_iterator_t fun, void *data)
     dico_iterator_t itr = xdico_list_iterator(database_list);
     dicod_database_t *db;
     int rc = 0;
-    
+
     for (db = dico_iterator_first(itr); rc == 0 && db;
 	 db = dico_iterator_next(itr)) {
-	if (database_visible_p(db)) 
+	if (database_is_visible(db))
 	    rc = fun(db, data);
     }
     dico_iterator_destroy(&itr);
@@ -1439,7 +1571,7 @@ database_remove_dependent(dicod_module_instance_t *inst)
 	if (dp->instance == inst) {
 	    dico_log(L_NOTICE, 0, _("removing database %s"), dp->name);
 	    dico_iterator_remove_current(itr, NULL);
-	    dicod_database_free(dp); 
+	    dicod_database_free(dp);
 	}
     }
     dico_iterator_destroy(&itr);
@@ -1459,28 +1591,28 @@ dicod_database_free(dicod_database_t *dp)
 
 void
 syslog_log_printer(int lvl, int exitcode, int errcode,
-                   const char *fmt, va_list ap)
+		   const char *fmt, va_list ap)
 {
     char *s;
     int prio = LOG_INFO;
     static struct {
-        char *prefix;
-        int priority;
+	char *prefix;
+	int priority;
     } loglevels[] = {
-        { "Debug",  LOG_DEBUG },
-        { "Info",   LOG_INFO },
-        { "Notice", LOG_NOTICE },
-        { "Warning",LOG_WARNING },
-        { "Error",  LOG_ERR },
-        { "CRIT",   LOG_CRIT },
-        { "ALERT",  LOG_ALERT },
-        { "EMERG",  LOG_EMERG },
+	{ "Debug",  LOG_DEBUG },
+	{ "Info",   LOG_INFO },
+	{ "Notice", LOG_NOTICE },
+	{ "Warning",LOG_WARNING },
+	{ "Error",  LOG_ERR },
+	{ "CRIT",   LOG_CRIT },
+	{ "ALERT",  LOG_ALERT },
+	{ "EMERG",  LOG_EMERG },
     };
     char buf[512];
     int level = 0;
 
     if (lvl & L_CONS)
-        _dico_stderr_log_printer(lvl, exitcode, errcode, fmt, ap);
+	_dico_stderr_log_printer(lvl, exitcode, errcode, fmt, ap);
 
     s    = loglevels[lvl & L_MASK].prefix;
     prio = loglevels[lvl & L_MASK].priority;
@@ -1489,8 +1621,8 @@ syslog_log_printer(int lvl, int exitcode, int errcode,
 	level += snprintf(buf + level, sizeof(buf) - level, "%s: ", s);
     level += vsnprintf(buf + level, sizeof(buf) - level, fmt, ap);
     if (errcode)
-        level += snprintf(buf + level, sizeof(buf) - level, ": %s",
-                          strerror(errcode));
+	level += snprintf(buf + level, sizeof(buf) - level, ": %s",
+			  strerror(errcode));
     syslog(prio, "%s", buf);
 }
 
@@ -1530,14 +1662,14 @@ _print_diag(grecs_locus_t const *locus, int err, int errcode, const char *msg)
 }
 
 void
-dicod_log_setup()
+dicod_log_setup(void)
 {
     if (!log_to_stderr) {
 	openlog(log_tag, LOG_PID, log_facility);
 	dico_set_log_printer(syslog_log_printer);
     }
     grecs_print_diag_fun = _print_diag;
-}    
+}
 
 /* When requested restart by a SIGHUP, the daemon first starts
    a copy of itself with the `--lint' option to verify
@@ -1547,7 +1679,7 @@ dicod_log_setup()
    __DICTD_LOGGING__ environment variable. */
 
 void
-dicod_log_encode_envar()
+dicod_log_encode_envar(void)
 {
     char *p;
     asprintf(&p, "%d:%d:%s", log_facility, log_print_severity, log_tag);
@@ -1555,7 +1687,7 @@ dicod_log_encode_envar()
 }
 
 void
-dicod_log_pre_setup()
+dicod_log_pre_setup(void)
 {
     char *str = getenv(DICTD_LOGGING_ENVAR);
     if (str) {
@@ -1563,7 +1695,7 @@ dicod_log_pre_setup()
 	log_facility = strtoul(str, &p, 10);
 	if (*p == ':') {
 	    log_print_severity = strtoul(p + 1, &p, 10);
-	    if (*p == ':') 
+	    if (*p == ':')
 		log_tag = p + 1;
 	}
 	log_to_stderr = 0;
@@ -1584,10 +1716,9 @@ apply_conf_override(struct dicod_conf_override *ovr)
     if (ovr->transcript >= 0)
 	transcript = ovr->transcript;
 }
-
 
 static void
-udb_init()
+udb_init(void)
 {
     dico_udb_define(&text_udb_def);
 }
@@ -1624,13 +1755,17 @@ main(int argc, char **argv)
     config_init();
     init_conf_override(&ovr);
 
-    get_options(argc, argv, &ovr);
+    get_options(&argc, &argv, &ovr);
     if (!config_lint_option)
 	dicod_log_setup();
-    
+
     if (mode == MODE_PREPROC)
 	exit(grecs_preproc_run(config_file, grecs_preprocessor) ?
 	      EX_CONFIG : 0);
+    else if (mode == MODE_TEST) {
+	dicod_loader_init();
+	exit(dicod_module_test(argc, argv));
+    }
 
     config_parse();
     /* Logging settings may have been changed by the config, so setup
@@ -1638,11 +1773,11 @@ main(int argc, char **argv)
     dicod_log_setup();
 
     apply_conf_override(&ovr);
-    
+
     compile_access_log();
 
     dicod_loader_init();
-    
+
     begin_timing("server");
     dicod_server_init();
     init_user_db();
@@ -1656,14 +1791,14 @@ main(int argc, char **argv)
 	exit(EX_SOFTWARE);
     if (config_lint_option)
 	exit(EX_OK);
-	
+
     markup_flush_capa();
 
     switch (mode) {
     case MODE_DAEMON:
 	dicod_server(argc, argv);
 	break;
-	
+
     case MODE_INETD:
 	if (dicod_inetd())
 	    rc = EX_UNAVAILABLE;
