@@ -1,6 +1,4 @@
-#ifdef HAVE_CONFIG_H
-# include <config.h>
-#endif
+#include <config.h>
 #include "dico.h"
 
 static const unsigned start[] = {
