@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico.
-   Copyright (C) 2008-2018 Sergey Poznyakoff
+   Copyright (C) 2008-2020 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -14,9 +14,7 @@
    You should have received a copy of the GNU General Public License
    along with GNU Dico.  If not, see <http://www.gnu.org/licenses/>. */
 
-#ifdef HAVE_CONFIG_H
-# include <config.h>
-#endif
+#include <config.h>
 #include <dico.h>
 #include <appi18n.h>
 #include <wordsplit.h>
@@ -281,7 +279,7 @@ _dico_ldap_search(struct _dico_ldap_handle *lp, const char *filter_pat,
     if (filter_pat) {
 	filter = _dico_ldap_expand_user(filter_pat, user);
 	if (!filter) {
-	    dico_log(L_ERR, ENOMEM, "_dico_ldap_expand_user");
+            DICO_LOG_MEMERR();
 	    return NULL;
 	}
     } else
